/*
 * devglue.h
 *
 * Copyright (c) 2016 Socionext Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2  of
 * the License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef __DEVGLUE_H
#define __DEVGLUE_H

#include <linux/device.h>
#include <linux/file.h>
#include <linux/interrupt.h>

/* definitions */
#define DEVGLUE_MAX_DEVNUM	1024
#define DEVGLUE_MAX_NAMELEN	16
#define DEVGLUE_MAX_PATHLEN	256

/* memory attribute types */
enum devglue_memory_type {
	DEVGLUE_MEMORY_UC = 1,	/* uncached */
	DEVGLUE_MEMORY_WC,	/* write-combined */
};

/* API prototypes */
void __iomem *devglue_ioremap(const char *, enum devglue_memory_type);
void devglue_iounmap(void __iomem *);
int devglue_reg_get_byname(const char *, phys_addr_t *, size_t *);

int devglue_request_irq(unsigned int, irq_handler_t, unsigned long, const char *, void *);
int devglue_request_irq_byname(unsigned int *, irq_handler_t, unsigned long, const char *, void *);
void devglue_free_irq(unsigned int, void *);
void devglue_enable_irq(unsigned int);
void devglue_enable_irq_byname(const char *);
void devglue_disable_irq(unsigned int);
void devglue_disable_irq_byname(const char *);
int devglue_irq_get(unsigned int);
int devglue_irq_get_byname(const char *);

struct device **devglue_alloc_device(const char *, unsigned int, unsigned int, const struct file_operations *);
void devglue_free_device(struct device **);

#endif /* __DEVGLUE_H */
