/*
 * dgsample_core.c
 * Sample driver for devglue
 *
 * Copyright (c) 2016 Socionext Inc.
 */
#include <linux/module.h>
#include <linux/device.h>	// device_*, class_*
#include <linux/slab.h>		// kzalloc
#include <linux/fs.h>		// inode
#include <linux/dma-mapping.h>	// dma_*
#include <linux/interrupt.h>

#include <asm/devglue.h>

#define DEV_NAME	"dgsample"
#define DEV_PATH	"devglue/" DEV_NAME
#define DEV_NUM		4
#define DEV_MAJOR	999

#define DEV_ID	((void*)-1)

static struct device **_dev_array;

/* device file operations */

static int dgsample_fops_open(struct inode *inode, struct file *filp)
{
	int minor = MINOR(inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static int dgsample_fops_release(struct inode *inode, struct file *filp)
{
	int minor = MINOR(inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static ssize_t dgsample_fops_read(struct file *flip, char *buf, size_t count, loff_t *pos)
{
	int minor = MINOR(flip->f_inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static ssize_t dgsample_fops_write(struct file *filp, const char *buf, size_t count, loff_t *pos)
{
	int minor = MINOR(filp->f_inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static long dgsample_fops_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long param)
{
	int minor = MINOR(filp->f_inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static long dgsample_fops_compat_ioctl(struct file *filp, unsigned int cmd, unsigned long param)
{
	int minor = MINOR(filp->f_inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static int dgsample_fops_mmap(struct file *filp, struct vm_area_struct *vma)
{
	int minor = MINOR(filp->f_inode->i_rdev);
	struct device *dev;

	dev = _dev_array[minor];
	dev_info(dev, "%s %d\n", __func__,minor);
	return 0;
}

static struct file_operations dgsample_fops = {
	.owner = THIS_MODULE,
	.open           = dgsample_fops_open,
	.release        = dgsample_fops_release,
	.read           = dgsample_fops_read,
	.write          = dgsample_fops_write,
	.unlocked_ioctl = dgsample_fops_unlocked_ioctl,
	.compat_ioctl   = dgsample_fops_compat_ioctl,
	.mmap           = dgsample_fops_mmap,
};

/* sample function for DMA-API (not called) */
static void dgsample_flush_range(const void *vptr, size_t size)
{
	struct device *dev = _dev_array[0];
	dma_addr_t dh;

	dh = dma_map_single(dev, vptr, size, DMA_TO_DEVICE);
	{
		dma_sync_single_for_cpu(dev, dh, size, DMA_TO_DEVICE);
		/*
		 * write from CPU here
		 */
		dma_sync_single_for_device(dev, dh, size, DMA_TO_DEVICE);
		/*
		 * read from DEVICE here
		 */
	}
	dma_unmap_single(dev, dh, size, DMA_TO_DEVICE);

	return;
}

/* irq handler */
static irqreturn_t dgsample_handler(int irq, void *dev_id)
{
	return IRQ_HANDLED;
}

/* initialize module */
static int __init dgsample_init_module(void)
{
	int ret;

	_dev_array = devglue_alloc_device(DEV_PATH, DEV_NUM, DEV_MAJOR, &dgsample_fops);
	if (IS_ERR(_dev_array)) {
		ret = PTR_ERR(_dev_array);
		goto err_dev;
	}

	/* interrupt */
	if ((ret = devglue_request_irq(246, dgsample_handler, IRQF_SHARED, "foo0", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(247, dgsample_handler, IRQF_SHARED, "foo1", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(248, dgsample_handler, IRQF_SHARED, "foo2", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(249, dgsample_handler, IRQF_SHARED, "foo3", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(250, dgsample_handler, IRQF_SHARED, "foo4", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(251, dgsample_handler, IRQF_SHARED, "bar0", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(252, dgsample_handler, IRQF_SHARED, "bar1", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(253, dgsample_handler, IRQF_SHARED, "bar2", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(254, dgsample_handler, IRQF_SHARED, "bar3", DEV_ID)) < 0) goto err_irq;
	if ((ret = devglue_request_irq(255, dgsample_handler, IRQF_SHARED, "bar4", DEV_ID)) < 0) goto err_irq;

	pr_info("%s: loaded.\n", DEV_NAME);
	return 0;

err_irq:
	devglue_free_irq(246, DEV_ID);
	devglue_free_irq(247, DEV_ID);
	devglue_free_irq(248, DEV_ID);
	devglue_free_irq(249, DEV_ID);
	devglue_free_irq(250, DEV_ID);
	devglue_free_irq(251, DEV_ID);
	devglue_free_irq(252, DEV_ID);
	devglue_free_irq(253, DEV_ID);
	devglue_free_irq(254, DEV_ID);
	devglue_free_irq(255, DEV_ID);

err_dev:

	pr_info("%s: error %d\n", DEV_NAME, ret);
	return ret;
}

/* clean up module */
static void __exit dgsample_cleanup_module(void)
{
	struct device **devp = _dev_array;

	devglue_free_irq(246, DEV_ID);
	devglue_free_irq(247, DEV_ID);
	devglue_free_irq(248, DEV_ID);
	devglue_free_irq(249, DEV_ID);
	devglue_free_irq(250, DEV_ID);
	devglue_free_irq(251, DEV_ID);
	devglue_free_irq(252, DEV_ID);
	devglue_free_irq(253, DEV_ID);
	devglue_free_irq(254, DEV_ID);
	devglue_free_irq(255, DEV_ID);

	devglue_free_device(devp);

	pr_info("%s: removed.\n", DEV_NAME);
	return;
}

module_init(dgsample_init_module);
module_exit(dgsample_cleanup_module);

MODULE_AUTHOR("Socionext Inc.");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("dgsample: sample driver for devglue");
